/*
*
*
* Based on "Applied Cryptography" book
* Dev:
*    Me
*/

#include <linux/init.h>
#include <linux/module.h>
#include <linux/scatterlist.h>
#include <linux/crypto.h>
#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <crypto/gost28147-89.h>

int gost_expand_key(struct gost_ctx *CTX, const u8 *key, unsigned int keylen)
{
	if(unlikely(keylen != GOST_MAX_KEY_SIZE))
		goto gost_expand_keysize_err;

	memset(CTX, 0, sizeof(struct gost_ctx));

	/* Copy of the cipher key */
	memcpy(CTX->key_enc, key, keylen);
	if(IS_ERR(CTX->key_dec))
		goto gost_expand_copy_err;

	/* Copy of the decipher key */
	memcpy(CTX->key_dec, key, keylen);
	if(IS_ERR(CTX->key_dec))
		goto gost_expand_copy_err;

	CTX->key_length = keylen;
	CTX->sbox = &gost_sboxes[0];

	return 0;

 gost_expand_keysize_err:
	pr_err("An error occurred with the key size, key size is %u,"
	       "expected size is %u\n", keylen, GOST_MAX_KEY_SIZE);
	return -1;

 gost_expand_copy_err:
	pr_err("An error occurred during the copy of the structure\n");
	return -2;
}

inline void split_in_two(const u8 *src, u32 block[2])
{
	/* MSB */
	block[0] = *(u32 *) src;
	/* LSB */
	block[1] = *(u32 *) (src + sizeof(u32));
}
inline void merge_in_one(u32 block[2], u8 *dst)
{
	dst[0] = block[0]	& 0xFF;
	dst[1] = block[0] >> 8	& 0xFF;
	dst[2] = block[0] >> 16 & 0xFF;
	dst[3] = block[0] >> 24 & 0xFF;
	dst[4] = block[1]	& 0xFF;
	dst[5] = block[1] >> 8	& 0xFF;
	dst[6] = block[1] >> 16 & 0xFF;
	dst[7] = block[1] >> 24 & 0xFF;
}
inline u32 transform(u32 cm1, struct gost_ctx *CTX){

	cm1 = (uint32_t) (CTX->sbox->sbox[7][(cm1>>28) & 0xF]<< 28 |
			  CTX->sbox->sbox[6][cm1>>24 & 0xF] << 24 |
			  CTX->sbox->sbox[5][cm1>>20 & 0xF] << 20 |
			  CTX->sbox->sbox[4][cm1>>16 & 0xF] << 16 |
			  CTX->sbox->sbox[3][cm1>>12 & 0xF] << 12 |
			  CTX->sbox->sbox[2][cm1>> 8 & 0xF] <<  8 |
			  CTX->sbox->sbox[1][cm1>> 4 & 0xF] <<  4 |
			  CTX->sbox->sbox[0][cm1     & 0xF]);
	return (cm1 << 11) | (cm1 >> 21);
}
int gost_setkey(struct crypto_tfm * tfm, const u8 * key,
		unsigned int keylen)
{
	int ret;
	struct gost_ctx *CTX = crypto_tfm_ctx(tfm);

	ret = gost_expand_key(CTX, key, keylen);
	if(unlikely(ret != 0))
		goto gost_expand_err;

	return 0;

 gost_expand_err:
	pr_err("An error occurred during initialization of the key\n");
	return -1;

}
EXPORT_SYMBOL(gost_setkey);

void gost_encrypt(struct crypto_tfm * tfm,
		  u8 * dst,
		  const u8 * src)
{
	u32 in[2], out[2];
	register u32 n1, n2; 	/* see rfc5830 */
	struct gost_ctx *CTX = crypto_tfm_ctx(tfm);

	split_in_two(src, in);



	n1 = in[1];		/* R_{i-1} (lsb)*/
	n2 = in[0];		/* L_{i-1} (msb) */

	n2 ^= transform(n1+CTX->key_enc[0], CTX);/* n2 = R_{i}; n1 = R_{i-1} = L_{i} */
	n1 ^= transform(n2+CTX->key_enc[1], CTX);/* n1 = R{i+1}; n2 = R_{i} =  L_{i+1} */
	n2 ^= transform(n1+CTX->key_enc[2], CTX);
	n1 ^= transform(n2+CTX->key_enc[3], CTX);
	n2 ^= transform(n1+CTX->key_enc[4], CTX);
	n1 ^= transform(n2+CTX->key_enc[5], CTX);
	n2 ^= transform(n1+CTX->key_enc[6], CTX);
	n1 ^= transform(n2+CTX->key_enc[7], CTX);

	n2 ^= transform(n1+CTX->key_enc[0], CTX);
	n1 ^= transform(n2+CTX->key_enc[1], CTX);
	n2 ^= transform(n1+CTX->key_enc[2], CTX);
	n1 ^= transform(n2+CTX->key_enc[3], CTX);
	n2 ^= transform(n1+CTX->key_enc[4], CTX);
	n1 ^= transform(n2+CTX->key_enc[5], CTX);
	n2 ^= transform(n1+CTX->key_enc[6], CTX);
	n1 ^= transform(n2+CTX->key_enc[7], CTX);

	n2 ^= transform(n1+CTX->key_enc[0], CTX);
	n1 ^= transform(n2+CTX->key_enc[1], CTX);
	n2 ^= transform(n1+CTX->key_enc[2], CTX);
	n1 ^= transform(n2+CTX->key_enc[3], CTX);
	n2 ^= transform(n1+CTX->key_enc[4], CTX);
	n1 ^= transform(n2+CTX->key_enc[5], CTX);
	n2 ^= transform(n1+CTX->key_enc[6], CTX);
	n1 ^= transform(n2+CTX->key_enc[7], CTX);

	n2 ^= transform(n1+CTX->key_enc[7], CTX);
	n1 ^= transform(n2+CTX->key_enc[6], CTX);
	n2 ^= transform(n1+CTX->key_enc[5], CTX);
	n1 ^= transform(n2+CTX->key_enc[4], CTX);
	n2 ^= transform(n1+CTX->key_enc[3], CTX);
	n1 ^= transform(n2+CTX->key_enc[2], CTX);
	n2 ^= transform(n1+CTX->key_enc[1], CTX);
	n1 ^= transform(n2+CTX->key_enc[0], CTX);

	/* In the RFC order - N1, then N2 */
	out[0] = n2;
	out[1] = n1;
	merge_in_one(out, dst);

}
EXPORT_SYMBOL(gost_encrypt);
void gost_decrypt(struct crypto_tfm * tfm,
		  u8 * dst,
		  const u8 * src)
{

	u32 in[2], out[2];
	register u32 n1, n2; 	/* see rfc5830 */
	struct gost_ctx *CTX = crypto_tfm_ctx(tfm);

	split_in_two(src, in);

	n1 = in[0];		/* R_{i-1} (lsb)*/
	n2 = in[1];		/* L_{i-1} (msb) */

	n2 ^= transform(n1+CTX->key_dec[0], CTX); /* n2 = R_{i}; n1 = R_{i-1} = L_{i} */
	n1 ^= transform(n2+CTX->key_dec[1], CTX); /* n1 = R_{i+1}; n2 = R_{i} =  L_{i+1} */
	n2 ^= transform(n1+CTX->key_dec[2], CTX);
	n1 ^= transform(n2+CTX->key_dec[3], CTX);
	n2 ^= transform(n1+CTX->key_dec[4], CTX);
	n1 ^= transform(n2+CTX->key_dec[5], CTX);
	n2 ^= transform(n1+CTX->key_dec[6], CTX);
	n1 ^= transform(n2+CTX->key_dec[7], CTX);

	n2 ^= transform(n1+CTX->key_dec[7], CTX);
	n1 ^= transform(n2+CTX->key_dec[6], CTX);
	n2 ^= transform(n1+CTX->key_dec[5], CTX);
	n1 ^= transform(n2+CTX->key_dec[4], CTX);
	n2 ^= transform(n1+CTX->key_dec[3], CTX);
	n1 ^= transform(n2+CTX->key_dec[2], CTX);
	n2 ^= transform(n1+CTX->key_dec[1], CTX);
	n1 ^= transform(n2+CTX->key_dec[0], CTX);


	n2 ^= transform(n1+CTX->key_dec[7], CTX);
	n1 ^= transform(n2+CTX->key_dec[6], CTX);
	n2 ^= transform(n1+CTX->key_dec[5], CTX);
	n1 ^= transform(n2+CTX->key_dec[4], CTX);
	n2 ^= transform(n1+CTX->key_dec[3], CTX);
	n1 ^= transform(n2+CTX->key_dec[2], CTX);
	n2 ^= transform(n1+CTX->key_dec[1], CTX);
	n1 ^= transform(n2+CTX->key_dec[0], CTX);


	n2 ^= transform(n1+CTX->key_dec[7], CTX);
	n1 ^= transform(n2+CTX->key_dec[6], CTX);
	n2 ^= transform(n1+CTX->key_dec[5], CTX);
	n1 ^= transform(n2+CTX->key_dec[4], CTX);
	n2 ^= transform(n1+CTX->key_dec[3], CTX);
	n1 ^= transform(n2+CTX->key_dec[2], CTX);
	n2 ^= transform(n1+CTX->key_dec[1], CTX);
	n1 ^= transform(n2+CTX->key_dec[0], CTX);

	out[0] = n1;
	out[1] = n2;

	merge_in_one(out, dst);

}
EXPORT_SYMBOL(gost_decrypt);
static struct crypto_alg gost_alg = {
	.cra_name		=	"gost",
	.cra_driver_name	=	"gost-generic",
	.cra_priority	        =       100,
	.cra_flags		=	CRYPTO_ALG_TYPE_CIPHER,
	.cra_blocksize		=	GOST_BLOCK_SIZE,
	.cra_ctxsize		=	sizeof(struct gost_ctx),
	.cra_alignmask		=	3,
	.cra_module		=	THIS_MODULE,
	.cra_u			=	{
		.cipher = {
			.cia_min_keysize	=	GOST_MIN_KEY_SIZE,
			.cia_max_keysize	=	GOST_MAX_KEY_SIZE,
			.cia_setkey		=	gost_setkey,
			.cia_encrypt		=	gost_encrypt,
			.cia_decrypt		=	gost_decrypt
		}
	}
};

static int __init gost_init(void)
{
	pr_info("Initialization : GOST 28147-89 module\n");
	return crypto_register_alg(&gost_alg);
}

static void __exit gost_exit(void)
{
	pr_info("Exit : GOST 28147-89 module\n");
	crypto_unregister_alg(&gost_alg);
}
module_init(gost_init);
module_exit(gost_exit);

MODULE_DESCRIPTION("GOST 28147-89 cipher algorithm");
MODULE_LICENSE("GPL");
MODULE_ALIAS_CRYPTO("gost28147-89");
MODULE_ALIAS_CRYPTO("gost89");
